@@ -6,4 +6,4 @@ import lit.formats
 config.name = 'llvm-mutate'
 config.test_format = lit.formats.ShTest(True)
 
-config.suffixed = ['.ll']
\ No newline at end of file
+config.suffixes = ['.ll']
